<?php

namespace Drupal\commercetools_content\Form;

use Drupal\commercetools\CommercetoolsService;
use Drupal\commercetools_content\Service\CommercetoolsContentComponents;
use Drupal\Core\Form\BaseFormIdInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides catalog filters.
 *
 * @internal
 */
class CatalogFiltersForm extends FormBase implements BaseFormIdInterface {

  /**
   * {@inheritdoc}
   */
  public function getBaseFormId() {
    return 'commercetools_content_catalog_filters_form';
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    // Try to fix an error with several identical forms on the same page.
    // @todo Remove after https://www.drupal.org/project/drupal/issues/2821852
    static $index = 0;
    return $this->getBaseFormId() . $index++;
  }

  /**
   * The Commercetools service.
   *
   * @var \Drupal\commercetools\CommercetoolsService
   */
  protected CommercetoolsService $ct;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected RendererInterface $renderer;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->ct = $container->get('commercetools');
    $instance->renderer = $container->get('renderer');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $install_state = NULL) {
    [$searchResultFacetData, $enabledFilters, $activeFilters, $productListIndex] = $form_state->getBuildInfo()['args'];
    $paramName = CommercetoolsContentComponents::getParamNameByIndex('filters', $productListIndex);

    $form['#attributes']['data-product-list-index'] = $productListIndex;
    $form['#attributes']['class'][] = 'commercetools-content-auto-submit';
    $form['#attributes']['class'][] = 'commercetools-filters-form';
    $form[$paramName] = [
      '#type' => 'container',
      '#tree' => TRUE,
    ];

    // Build filters.
    foreach ($enabledFilters as $filter) {
      $path = $filter['path'];
      switch ($filter['widget_type']) {
        case 'facet':
        case 'facet_count':
          $form[$paramName][$path] = $this->buildFacetFilter($filter, $searchResultFacetData, $activeFilters);
          break;

        case 'textfield':
        case 'checkbox':
          $form[$paramName][$path] = $this->buildSimpleFilter($filter, $activeFilters);
          break;

        case 'separator':
          $form[$paramName][$path] = $this->buildSeparator($filter);
          break;
      }
    }

    $form['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Apply'),
    ];

    return $form;
  }

  /**
   * Build faced filter.
   */
  protected function buildFacetFilter(array $filter, array $facetResults = [], array $activeFilters = []) {
    $label = [
      '#theme' => 'commercetools_filter_label',
      '#label' => $filter['label'],
    ];
    $element = [
      '#title' => $this->renderer->render($label),
      '#type' => 'fieldset',
    ];

    foreach ($facetResults as $facetResult) {
      if ($facetResult['facet'] !== $filter['path']) {
        continue;
      }
      foreach ($facetResult['value']['terms'] as $term) {
        $labelRender = [
          '#theme' => 'commercetools_filter_option',
          '#label' => $term['label'] ?? $term['term'],
          '#count' => $filter['widget_type'] === 'facet_count' ? $term['productCount'] : NULL,
        ];
        $element[] = [
          '#type' => 'checkbox',
          '#title' => $this->renderer->render($labelRender),
          '#return_value' => $term['term'],
          '#default_value' => in_array($term['term'], $activeFilters[$facetResult['facet']] ?? []),
          '#attributes' => [
            'class' => ['facet-filter'],
            'data-product-attribute-key' => $filter['type'] === 'attribute' ? $filter['attributeKey'] : NULL,
          ],
        ];
      }
    }

    return $element;
  }

  /**
   * Build simple filter.
   */
  protected function buildSimpleFilter($filter, array $activeFilters = []) {
    $group_label = [
      '#theme' => 'commercetools_filter_label',
      '#label' => $filter['label'],
    ];

    return [
      '#title' => $this->renderer->renderRoot($group_label),
      '#type' => $filter['widget_type'],
      '#default_value' => $activeFilters[$filter['path']] ?? NULL,
    ];
  }

  /**
   * Build simple filter.
   */
  protected function buildSeparator($separator) {
    return [
      '#type' => 'html_tag',
      '#tag' => 'h3',
      '#value' => $separator['label'],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $query = $this->getRequest()->query->all();
    // Reset page after applying new facets.
    unset($query['page']);

    [, $enabledFilters, , $productListIndex, $targetPage] = $form_state->getBuildInfo()['args'];
    $paramName = CommercetoolsContentComponents::getParamNameByIndex('filters', $productListIndex);

    $query = [];
    $filtersValues = $form_state->getValue($paramName);
    foreach ($enabledFilters as $filter) {
      $filterValue = $filtersValues[$filter['path']] ?? NULL;
      $filterValue = is_array($filterValue) ? array_filter($filterValue) : $filterValue;

      if (empty($filterValue)) {
        unset($query[$paramName][$filter['path']]);
      }
      else {
        $query[$paramName][$filter['path']] = $filterValue;
      }
    }

    $targetPage ??= Url::fromRoute('<current>');
    $targetPage->setOption('query', array_filter($query));
    $form_state->setRedirectUrl($targetPage);
  }

  /**
   * Convert camel case to kebab case.
   */
  protected function camelToKebab(string $string): string {
    return strtolower(preg_replace('/[A-Z]/', '-$0', $string));
  }

  /**
   * Initialize the form state and the entity before the first form build.
   */
  protected function init(FormStateInterface $form_state) {

  }

}
