class FiltersForm extends HTMLElement {
  connectedCallback() {
    this.paramName = window.commercetools.getParamNameByIndex(
      'filters',
      this.productListIndex,
    );

    const form = document.createElement('form');
    form.classList.add('filters-form', 'commercetools-filters-form');
    form.innerHTML = '';
    if (this.filterSubmitHandler) {
      form.addEventListener('submit', this.filterSubmitHandler);
    }

    try {
      this.renderFilters(form);
    } catch (error) {
      new Drupal.Message().add(Drupal.t('Failed to load filters.'), {
        type: 'error',
      });
      throw error;
    }
    this.form = form;
    this.appendChild(form);
  }

  renderFilters(form) {
    if (this.isLoading) {
      form.insertAdjacentHTML('beforeend', FiltersForm.placeholder());
      return;
    }
    if (!this.filters.length) {
      return;
    }

    this.filters.forEach((filterObj) => {
      let filter;
      switch (filterObj.widget_type) {
        case 'facet':
        case 'facet_count':
          filter = this.buildFacetFilter(filterObj);
          break;

        case 'textfield':
        case 'checkbox':
          filter = this.buildSimpleFilter(filterObj);
          break;

        case 'separator':
          filter = FiltersForm.buildSeparator(filterObj);
          break;
      }
      if (filter) {
        form.append(filter);
      }
    });

    const applyButton = document.createElement('input');
    applyButton.value = Drupal.t('Apply');
    applyButton.type = 'submit';
    applyButton.classList.add(
      'btn-lg',
      'button',
      'js-form-submit',
      'form-submit',
      'btn',
      'btn-primary',
    );
    form.append(applyButton);
  }

  static buildSeparator(filterObj) {
    const separator = document.createElement('h3');
    separator.innerHTML = filterObj.label;
    return separator;
  }

  buildSimpleFilter(filterObj) {
    const fClass = filterObj.path
      .replace(/\./g, '-')
      .replace(/[A-Z]/g, '-$&')
      .toLowerCase();
    const id = `${this.paramName}-${fClass}`;
    const type = filterObj.widget_type;

    const wrapper = document.createElement('div');
    wrapper.className = `js-form-item js-form-type-${type} ${type} mb-3 js-form-item-${id}`;

    const label = document.createElement('label');
    label.className = 'form-check-label';
    label.setAttribute('for', `edit-${id}`);
    label.innerHTML = `<span>${filterObj.label}</span>`;

    const inputName = `${this.paramName}[${filterObj.path}]`;
    const input = document.createElement('input');
    input.name = inputName;
    input.id = `edit-${id}`;
    if (type === 'checkbox') {
      input.type = 'checkbox';
      input.value = '1';
      input.className = 'form-checkbox form-check-input';
      input.checked = this.activeFilters?.[filterObj.path]?.[0];
      wrapper.appendChild(input);
      wrapper.appendChild(label);
      wrapper.classList.add('form-check');
    } else {
      input.type = 'text';
      input.value = this.activeFilters?.[filterObj.path]?.[0] || '';
      input.size = 60;
      input.maxLength = 128;
      input.className = 'form-control';
      wrapper.appendChild(label);
      wrapper.appendChild(input);
    }

    return wrapper;
  }

  buildFacetFilter(filterObj) {
    const facetObj = this.response.facets.find(
      (facet) => facet.facet === filterObj.path,
    );

    if (!facetObj || !facetObj.value) {
      new Drupal.Message().add(Drupal.t('Invalid filter data.'), {
        type: 'warning',
      });
      return;
    }

    // Hide the empty facet.
    if (!facetObj.value.terms.length) {
      return;
    }

    const fieldset = document.createElement('fieldset');
    fieldset.classList.add(
      'fieldgroup',
      'form-composite',
      'form-item',
      'form-wrapper',
    );
    const legend = document.createElement('legend');
    legend.innerHTML = `<span class="fieldset-legend">${filterObj.label}</span>`;
    fieldset.appendChild(legend);

    facetObj.value.terms.forEach((termObj) => {
      const { term, productCount } = termObj;
      const id = `${facetObj.facet}-${term}`.replace(/\W/g, '-');

      const checkbox = document.createElement('input');
      checkbox.type = 'checkbox';
      checkbox.name = `${this.paramName}[${facetObj.facet}][]`;
      checkbox.value = term;
      checkbox.checked = this.activeFilters[facetObj.facet]?.includes(term);
      checkbox.setAttribute('id', id);
      checkbox.setAttribute(
        'data-product-attribute-key',
        filterObj.attributeKey,
      );
      checkbox.classList.add('facet-filter', 'form-check-input');

      const label = document.createElement('label');
      label.htmlFor = id;
      label.classList.add('option');
      label.innerHTML = `<span class="text-truncate">${term}</span>`;
      if (filterObj.widget_type === 'facet_count') {
        label.innerHTML += `<span className="float-end">(${productCount})</span>`;
      }

      const wrapper = document.createElement('div');
      wrapper.className = `js-form-item js-form-type-checkbox checkbox mb-3 js-form-item-${id} form-check`;

      wrapper.appendChild(checkbox);
      wrapper.appendChild(label);
      fieldset.appendChild(wrapper);
    });
    return fieldset;
  }

  static placeholder() {
    return `
      <fieldset class="fieldgroup form-composite js-form-item form-item js-form-wrapper form-wrapper placeholderify">
        <legend><span class="fieldset-legend">Placeholder for the filter name</span></legend>
        <div class="fieldset-wrapper">
          ${Array.from({ length: 6 }, () => {
            return `
              <input type="checkbox" name="name" value="value" class="form-checkbox form-check-input">
              <label for="edit-name" class="option"><span class="ms-3 text-truncate">Filter checkbox</span> <span class="float-end">(10)</span></label>
              <hr />
            `;
          }).join('')}
        </div>
      </fieldset>
    `;
  }

  setLoadingState(state = null) {
    const stateClasses = {
      load: 'placeholderify',
      reload: 'reloadify',
    };

    this.form.classList.remove(...Object.values(stateClasses));
    const stateClass = stateClasses[state];
    if (state !== null && stateClass) {
      this.form.classList.add(stateClass);
    }
  }
}

customElements.define('ct-filters-form', FiltersForm);
