<?php

namespace Drupal\commercetools_demo;

use Drupal\commercetools\CommercetoolsConfiguration;
use Drupal\commercetools\CommercetoolsLocalization;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Component\Serialization\Json;
use Drupal\Component\Serialization\Yaml;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Routing\RouteBuilderInterface;
use Drupal\Core\State\StateInterface;
use Drupal\language\Entity\ConfigurableLanguage;

/**
 * Deploys demo Configuration presets.
 */
class DemoConfigurationDeployer {
  private const CONFIG_DIRECTORY_PATH = 'config' . DIRECTORY_SEPARATOR . 'demo';

  private const CREDENTIALS_DIRECTORY_PATH = 'fixtures' . DIRECTORY_SEPARATOR . 'demo_accounts';
  private const THEMES_CONFIG_DIRECTORY_PATH = 'fixtures' . DIRECTORY_SEPARATOR . 'themes_config';
  private const UI_COMPONENTS_DIRECTORY_PATH = 'fixtures' . DIRECTORY_SEPARATOR . 'ui_components';
  private const DEMO_COMPONENTS_DIRECTORY_PATH = 'fixtures' . DIRECTORY_SEPARATOR . 'demo_components';

  private const STATE_KEY_UI_DEPLOYMENT_PREFIX = 'commercetools_demo.ui_deployment.';

  /**
   * CommercetoolsCarts constructor.
   *
   * @param \Drupal\Core\Routing\RouteBuilderInterface $routerBuilder
   *   The router builder.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\State\StateInterface $state
   *   The Drupal state storage service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory.
   * @param \Drupal\commercetools\CommercetoolsConfiguration $ctConfig
   *   The commercetools configuration service.
   * @param \Drupal\commercetools\CommercetoolsLocalization $ctLocalization
   *   The commercetools localization service.
   */
  public function __construct(
    protected readonly RouteBuilderInterface $routerBuilder,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly StateInterface $state,
    protected readonly ConfigFactoryInterface $configFactory,
    protected readonly CommercetoolsConfiguration $ctConfig,
    protected readonly CommercetoolsLocalization $ctLocalization,
  ) {
  }

  /**
   * Provides a default theme machine name.
   *
   * @return string
   *   A default theme machine name.
   */
  public function getDefaultTheme(): string {
    return $this->configFactory
      ->get('system.theme')
      ->get('default');
  }

  /**
   * Provides a list of YAML encoded presets in a directory.
   *
   * @param string $directory
   *   The absolute path to a directory to seek in.
   *
   * @return array
   *   A list of suggested themes keyed by name.
   */
  protected function getPresets(string $directory): array {
    $presets = [];
    $files = scandir($directory);
    foreach ($files as $file) {
      if (!preg_match('#^(.+)\.yml#', $file, $matches)) {
        continue;
      }
      $name = $matches[1];
      $presets[$name] = (array) Yaml::decode(file_get_contents($directory . DIRECTORY_SEPARATOR . $file));
    }

    return $presets;
  }

  /**
   * Provides a list of suggested themes.
   *
   * @return array
   *   A list of suggested themes keyed by machine name.
   */
  public function getSuggestedThemes(): array {
    $themesDirectory = dirname(__DIR__) . DIRECTORY_SEPARATOR . self::THEMES_CONFIG_DIRECTORY_PATH;
    return $this->getPresets($themesDirectory);
  }

  /**
   * Provides the list of demo account credentials.
   *
   * @return array
   *   An array with available demo configurations.
   */
  public function getDemoAccounts(): array {
    static $accounts = [];
    if (empty($accounts)) {
      $accountsDirectory = dirname(__DIR__) . DIRECTORY_SEPARATOR . self::CREDENTIALS_DIRECTORY_PATH;
      $accounts = $this->getPresets($accountsDirectory);
    }
    return $accounts;
  }

  /**
   * Deploys demo account configuration.
   *
   * @param string $account
   *   The demo account ID.
   */
  public function deployDemoAccount(string $account): void {
    $this->unsetConnectionConfig();
    $demoAccount = $this->getDemoAccounts()[$account];
    $demoConfig = $demoAccount['config'];
    foreach ($demoConfig as $configName => $values) {
      $config = $this->configFactory->getEditable($configName);
      if ($config) {
        foreach ($values as $key => $value) {
          $config->set($key, $value);
        }
      }
      $config->save();
    }
  }

  /**
   * Cleans up connection configuration.
   */
  public function unsetConnectionConfig(): void {
    $config = $this->configFactory->getEditable(CommercetoolsConfiguration::CONFIGURATION_API);
    $keys = $this->ctConfig->listCredentialKeys();
    // API scope is also related to credentials.
    $keys[] = 'scope';
    foreach ($keys as $key) {
      $config->set($key, '');
    }
    $config->save();
  }

  /**
   * Provides the demo configuration options for the module.
   *
   * @param string $module
   *   The module name.
   *
   * @return array
   *   The config values.
   */
  public function getDemoConfig(string $module): array {
    $fileName = "{$module}.settings.yml";
    $filePath = dirname(__DIR__) . DIRECTORY_SEPARATOR . self::CONFIG_DIRECTORY_PATH . DIRECTORY_SEPARATOR . $fileName;
    return (array) Yaml::decode(file_get_contents($filePath));
  }

  /**
   * Overrides configuration with demo values for the module.
   *
   * @param string $module
   *   The module name.
   */
  public function setDemoConfig(string $module): void {
    $demoConfig = $this->getDemoConfig($module);
    $configName = "{$module}.settings";
    $config = $this->configFactory->getEditable($configName);
    foreach ($demoConfig as $key => $value) {
      $config->set($key, $value);
    }
    $config->save();
    // We need to rebuild the router to apply the catalog path changes.
    $this->routerBuilder->rebuild();
  }

  /**
   * Provides the list of demo UI components for a module.
   *
   * @param string $module
   *   The UI module name.
   *
   * @return array
   *   The list of UI components.
   */
  public function getDemoComponents(string $module): array {
    $directory = match ($module) {
      'commercetools_demo' => self::DEMO_COMPONENTS_DIRECTORY_PATH,
      default => self::UI_COMPONENTS_DIRECTORY_PATH,
    };
    $componentsDirectory = dirname(__DIR__) . DIRECTORY_SEPARATOR . $directory;
    $components = $this->getPresets($componentsDirectory);
    foreach ($components as &$component) {
      // Check a component affiliation.
      if (!empty($component['_commercetools_metadata'])) {
        $metadata = $component['_commercetools_metadata'];
        // Allow-list.
        if (!empty($metadata['target_modules']) && !in_array($module, $metadata['target_modules'])) {
          continue;
        }
        // Deny-list.
        if (!empty($metadata['ignore_modules']) && in_array($module, $metadata['ignore_modules'])) {
          continue;
        }
        unset($component['_commercetools_metadata']);
      }
    }
    return $components;
  }

  /**
   * Checks if deployment is done for the module.
   *
   * @param string $module
   *   The UI module name.
   *
   * @return bool
   *   A deployment status.
   */
  public function isDemoComponentsDeployedFor(string $module): bool {
    $deployed = $this->state->get(self::STATE_KEY_UI_DEPLOYMENT_PREFIX . $module);
    if (!$deployed) {
      return FALSE;
    }
    foreach ($deployed as $id => $uuid) {
      preg_match('/(?<type>.+)\.(?<name>.+)/', $id, $matches);
      try {
        $storage = $this->entityTypeManager->getStorage($matches['type']);
        /** @var \Drupal\Core\Entity\EntityInterface $entity */
        $entity = $storage->loadByProperties(['uuid' => $uuid]);
        if (!$entity) {
          return FALSE;
        }
      }
      catch (PluginNotFoundException $e) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Creates a new entity by type and data given.
   *
   * @param string $type
   *   The entity type ID.
   * @param array $values
   *   The entity properties as associative array.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   A new entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function deployEntity(string $type, array $values): EntityInterface {
    switch ($type) {
      case 'configurable_language':
        $entity = ConfigurableLanguage::createFromLangcode($values['id']);
        if (array_key_exists($values['id'], $this->ctLocalization->languages)) {
          $values['id'] .= '-ct';
        }
        foreach ($values as $key => $value) {
          $entity->set($key, $value);
        }
        break;

      default:
        $storage = $this->entityTypeManager->getStorage($type);
        $entity = $storage->create($values);
    }
    try {
      $entity->save();
    }
    catch (EntityStorageException $e) {
      if ($storage instanceof EntityStorageInterface) {
        $entityDuplicate = $storage->load($entity->id());
        $entityDuplicate->delete();
        $entity->save();
      }
    }
    return $entity;
  }

  /**
   * Collects all values for replacing in variables.
   *
   * @param string $module
   *   The UI module name.
   * @param array $variables
   *   The variables to replace in values as [KEY]. Optional.
   *
   * @return array
   *   The list of replacements keyed by variable name.
   */
  protected function prepareVariables(string $module, array $variables = []): array {
    $moduleSettings = $this->configFactory->get("{$module}.settings");
    $variables += [
      'module' => $module,
      'module_label' => match ($module) {
        'commercetools_demo' => 'commercetools Demo',
        'commercetools_content' => 'commercetools Content',
        'commercetools_decoupled' => 'commercetools Decoupled',
        default => $module,
      },
      'catalog_path' => $moduleSettings->get('catalog_path'),
      ...$this->getThemeVariables(),
    ];

    return $variables;
  }

  /**
   * Collects all values for replacing in variables.
   *
   * @param array $component
   *   The component properties as associative array.
   * @param string $module
   *   The UI module name.
   * @param array $variables
   *   The variables to replace in values as [KEY]. Optional.
   *
   * @return array
   *   The prepared component.
   */
  protected function prepareComponentConfig(array $component, string $module, array $variables = []): array {
    $component = $this->replaceValuesVariables(
      $component,
      $this->prepareVariables($module, $variables),
    );
    $component += [
      'dependencies' => [
        'enforced' => [
          'module' => [
            'commercetools_demo',
            $module,
          ],
        ],
      ],
    ];
    return $component;
  }

  /**
   * Deploys demo components.
   *
   * @param string $module
   *   The UI module name.
   * @param array $variables
   *   The variables to replace in values as [KEY]. Optional.
   */
  public function deployDemoComponents(string $module, array $variables = []): void {
    $components = $this->getDemoComponents($module);
    $deployed = $this->state->get(self::STATE_KEY_UI_DEPLOYMENT_PREFIX . $module, []);
    foreach ($components as $id => $config) {
      preg_match('/(?<type>.+)\.(?<name>.+)/', $id, $matches);
      $config = $this->prepareComponentConfig($config, $module, $variables);
      $entity = $this->deployEntity($matches['type'], $config);
      $deployed[$id] = $entity->uuid();
    }
    $this->state->set(self::STATE_KEY_UI_DEPLOYMENT_PREFIX . $module, $deployed);
  }

  /**
   * Removes demo UI components for a module.
   *
   * @param string $module
   *   The UI module name.
   */
  public function removeDemoComponents(string $module): void {
    $deployed = $this->state->get(self::STATE_KEY_UI_DEPLOYMENT_PREFIX . $module, []);
    foreach ($deployed as $id => $uuid) {
      preg_match('/(?<type>.+)\.(?<name>.+)/', $id, $matches);
      $storage = $this->entityTypeManager->getStorage($matches['type']);
      /** @var \Drupal\Core\Entity\EntityInterface $entity */
      $entity = current($storage->loadByProperties(['uuid' => $uuid]));
      if ($entity) {
        $entity->delete();
      }
    }
    $this->state->delete(self::STATE_KEY_UI_DEPLOYMENT_PREFIX . $module);
  }

  /**
   * Provides default variables for replacement.
   *
   * @return array
   *   The theme variables.
   */
  private function getThemeVariables(): array {
    $variables = [];
    $themeCurrent = $this->configFactory->get('system.theme')->get('default');
    $variables['theme'] = $themeCurrent;
    $themes = $this->getSuggestedThemes();
    $theme = $themes[$themeCurrent];
    foreach ($theme['regions'] as $key => $value) {
      $variables['region:' . $key] = $value;
    }
    return $variables;
  }

  /**
   * Replaces variables in values.
   *
   * @todo Consider token as alternative.
   *
   * @param array $values
   *   The values to check variables in.
   * @param array $variables
   *   The variables to replace in values as :[KEY].
   *
   * @return array
   *   The values with replacements.
   */
  private function replaceValuesVariables(array $values, array $variables): array {
    // $variables = $this->themeVariables() + $variables;
    $string = Json::encode($values);
    foreach ($variables as $key => $variable) {
      $search = "[$key]";
      $string = str_replace($search, $variable, $string);
    }
    return Json::decode($string);
  }

}
